/**
 ** A Virtual Host Monitor
 **
 ** Copyright 2009 Gary Wuertz gary@issiweb.com
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include "monitor.h"

#define REC_LEN     1024

using namespace std;

/**
 * Parse a log record. The 0th slot of the rec array is used as the parse space for
 * incremental parsing. Each new record is counted as a 'hit'
 */
void vhost::parseLogRecord(const std::map<int,std::string>* rmap,char *s, int ct)
{
   char              ch;
   string            key;
   map<int,string>   p = *rmap;

   this->meter += ct;
   while(ct--) {
    switch(ch=*s++) {
      case '"':
         this->iq = !this->iq;
         break;
      case ' ':
        if (this->iq || this->rec[0][0]=='[') {
            this->rec[0] += ch;
            break;
            }
         else if (!this->rec[0].length())
            break;
      case '\n':
   done:
         key = p[this->argn];
         switch(key[0]) {
            case 'h':
               this->rec[vhost::IP_SLOT]      = this->rec[0];
               break;
            case 'r':
               this->rec[vhost::REQUEST_SLOT] = this->rec[0];
            }
         if (ch=='\n') {
            this->argn = 0;
            this->hits++;
            }
         else this->argn++;
         this->rec[0] = "";
         break;
      case ']':
         this->rec[0] += ch;
         if (!this->iq && this->rec[0][0]=='[')
            goto done;
         break;
      default:
         this->rec[0] += ch;
      }
    }
}
/**
 * Clean the dirty bit
 */
bool vhost::setClean(const std::map<int,std::string>* rmap)
{
   char  buf[REC_LEN];
   const char *file = this->file.c_str();
   int   ct;

   int fh = open(file,O_RDONLY);
   if (fh!=-1) {
       if (lseek(fh,this->fsz,SEEK_SET)!=-1) {
         while((ct = read(fh,buf,sizeof(buf)))>0)
           this->parseLogRecord(rmap,buf,ct);
         this->fsz = lseek(fh,0,SEEK_CUR);
         }
      close(fh);
      return true;
      }
   return false;
}
/**
 * Save info for later processing
 */
bool vhost::setDirty(bool init)
{
   struct stat attrib;
   const char *file = this->file.c_str();

   if (stat(file,&attrib)==0) {
      this->ldtm  = attrib.st_mtime;
      this->dirty = init;
      if (!init)
         this->fsz = attrib.st_size;
      return true;
      }
   return false;
}
/**
 * Format vhost content using a sprintf like specification using the place holders:
 *
 *  %H  = hour
 *  %M  = minute
 *  %S  = second
 *  %Y  = year
 *  %d  = day
 *  %h  = remote host     - right aligned and truncated/padded to width
 *  %m  = month
 *  %o  = count (hits)
 *  %r  = request string  - left aligned and truncated/padded to width
 *  %v  = domain name     - left aligned and always padded to namelength
 */
void vhost::toString(topvhost *info)
{
   char buf[128];
   struct tm *ts = localtime(&this->ldtm);
   string token;
   int namelen = info->namelen;
   string::iterator fmt = info->rfmt.begin();
   char ch;

   info->tline = "";
   while( fmt != info->rfmt.end() )
      if ((ch = *fmt++)=='%') {
         token = ch;
         int svl = -1;
         int val = 0;
         if (strchr("-+#0 ",*fmt)!=NULL)           // flags
            token += *fmt++;
         while(strchr("0123456789.",*fmt)!=NULL)   // width - precision
            token += *fmt++;
         switch(*fmt++) {
            case 'H':   val = ts->tm_hour;   break;
            case 'M':   val = ts->tm_min;    break;
            case 'S':   val = ts->tm_sec;    break;
            case 'Y':   val = ts->tm_year;   break;
            case 'd':   val = ts->tm_mday;   break;
            case 'h':   svl = IP_SLOT;       break;
            case 'm':   val = ts->tm_mon+1;  break;
            case 'o':   val = this->hits;    break;
            case 'r':   svl = REQUEST_SLOT;  break;
            case 'v':   svl = PARSE_SLOT;    break;
            }
         if (svl==-1) {
            token += 'd';
            snprintf(buf,sizeof(buf),token.c_str(), val);
            info->tline += buf;
            }
         else {
            string str = svl==PARSE_SLOT? this->domain : this->rec[svl];
            token[0] = ' ';
            unsigned len = (unsigned) atoi(token.c_str());
            switch(svl) {
               case PARSE_SLOT:
                  len = namelen;
                  break;
               case IP_SLOT:
                  if (str.length()>len)
                     str = str.substr(str.length()-len);
                  break;
               case REQUEST_SLOT:
                  if (str.length()>len)
                     str = str.substr(0,len);
                  break;
               }
            while(str.length()<len)
               str += ' ';
            info->tline += str;
            }
         }
      else info->tline += ch;
}
